/*
  TROLL (formerly ResHiggs)
  Example of usage for DIS
*/

#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h> 
#include <sys/time.h> 
#include <getopt.h> 
#include <gsl/gsl_errno.h>

#include "./src/resummationQCD.hh"

using namespace std;



void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -p  --pdf-set <set>     specify PDFs being used (default \"NNPDF30_nnlo_as_0118\")" << endl
       << " -Q  --momentum <Q>      set transferred momentum Q [GeV]" << endl
       << " -x  --xBjorken <x>      set Bjorken x" << endl
       << " -T  --TargetMass <t>    set Target Mass (in units of the proton mass)" << endl
       << " -f  --fact <murat>      set muF/Q ratio" << endl
       << " -r  --ren <murat>       set muR/Q ratio" << endl
       << " -P  --prescription <p>  set resummation prescription: psi-soft (1), N-soft (2, default)" << endl
       << " -g  --g0exp             exponentiate constant function g0 or g0bar" << endl
    // << " -N  --Nsubtract <N0>    subtract constants in N=N0" << endl
       << endl;
  exit(0);
  return;
}




double Q;
double muRrat = 1;
double muFrat = 1;
bool g0exp = false;
int prescr = 2;
double protonMass = 0.938;

const string sorder[4] = {"LO", "NLO", "NNLO", "NNNLO"};
void read_arguments(int argc, char* argv[], double &q, double &x, string &set, double &Nsub, double &TM) {
  const char* const short_options = "hp:f:r:Q:x:T:P:gN:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "pdf-set", 1, NULL, 'p' },
					 { "fact", 1, NULL, 'f' },
					 { "ren", 1, NULL, 'r' },
					 { "momentum", 1, NULL, 'Q' },
					 { "xBjorken", 1, NULL, 'x' },
					 { "TargetMass", 1, NULL, 'T' },
					 { "prescription", 1, NULL, 'P' },
					 { "g0exp", 0, NULL, 'g' },
					 { "Nsubtract", 1, NULL, 'N' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'p':
      sset << optarg;
      set = sset.str();
      break;
    case 'f':
      muFrat = strtod(optarg, NULL);
      break;
    case 'r':
      muRrat = strtod(optarg, NULL);
      break;
    case 'Q':
      q = strtod(optarg, NULL);
      break;
    case 'x':
      x = strtod(optarg, NULL);
      break;
    case 'T':
      TM = strtod(optarg, NULL);
      TM *= protonMass;
      break;
    case 'P':
      prescr = strtol(optarg, NULL, 10);
      break;
    case 'g':
      g0exp = false;
      break;
    case 'N':
      Nsub = strtod(optarg, NULL);
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  //
  // messages
  //
  if(prescr==2 && g0exp==true)
    cout << endl << "\033[1m  WARNING:\033[0m "
	 << "\033[0;31m Using N-soft with g0 exponentiated: this is NOT the standard N-soft resummation prescription!\033[0m"
	 << endl << endl;
  //
  return;
}







int main (int argc, char* argv[]) {





  // default values (modified by runtime arguments)
  Q = 2;
  double x = 0.1;
  string PDFset = "NNPDF30_nnlo_as_0118";
  double Nsub = __builtin_nan("");
  double TM = 0;

  // read runtime arguments
  read_arguments(argc, argv, Q, x, PDFset, Nsub, TM);


  // Start time counter
  struct timeval t0, t1;
  gettimeofday(&t0,NULL);





  // *** Resummation part *** //
  //
  using namespace resumm;
  QCDResummation res(DISneutral, proton, PDFset);
  //
  res.SetLogCounting(primed);
  //
  // Now we select the prescription
  // The easiest way uses the name of the prescription, that can be
  //    "A-soft_n"   or   "psi-soft_n"   or   "N-soft"
  // where n is the order of the Altarelli-Parisi improvement, and can be 0, 1 or 2
  string PName[] = { "A-soft_2", "psi-soft_2", "N-soft", "Psi-soft_2" };
  res.SetPrescriptionByName(PName[prescr], g0exp);
  res.SetMPpars(14,1,0,0);
  res.SetTargetMass(TM);
  if(Nsub==Nsub) res.SetConstSubtraction(Nsub);  // please ignore this (under study)


  cout << "PDFset: " << PDFset << endl
       << "Renormalization scale muR = " << muRrat << " * Q" << endl
       << "Factorization   scale muF = " << muFrat << " * Q" << endl
       << "x = " << x << endl
       << "Q = " << Q << " GeV" << endl
       << "Target Mass = " << TM << " GeV" << endl
       << "prescription: " << PName[prescr] << endl
       << "exponentiate g0: " << (g0exp ? "yes" : "no") << endl
       << endl;
  //


  double xNLL   = res.res_xs(Q, Q/sqrt(x), muFrat, muRrat,   NLL,   NLO);
  double xNNLL  = res.res_xs(Q, Q/sqrt(x), muFrat, muRrat,  NNLL,  NNLO);
  double xNNNLL = res.res_xs(Q, Q/sqrt(x), muFrat, muRrat, NNNLL, NNNLO);


  cout << "Use this ouput to construct a resummed cross section as" << endl
       << "  sigma(NjLO+NjLL) = sigma(NjLO) + sigma(LO) * Delta_j K(NjLL)" << endl;
  if(TM>0) cout << "where sigma(NjLO) and sigma(LO) must contain Target Mass Corrections" << endl;
  cout << endl;

  cout << "Delta_1 K(NLL)   = " << xNLL << endl
       << "Delta_2 K(NNLL)  = " << xNNLL << endl
       << "Delta_3 K(NNNLL) = " << xNNNLL << endl
       << endl;

  
  // Stop time counter
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << endl << "Total time: " << t << "s" << endl;


  return 0;

}
