#include <cmath>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include <cuba.h>
#include "resummationQCD.hh"
#include "chebyshev.hh"
#include "lagrange.hh"
#include "QCD_EW.hh"
#include "math/special_functions.hh"
#include "math/integration.hh"
#include "math/struct_definitions.hh"


using namespace std;




namespace resumm {

  inline double zminoftau(double tau, double freezing) {
    return tau*(1-tau/2./freezing);  // this has zero derivative in tau=freezing
  }

  QCDResummation::~QCDResummation() {
    return;
  }


  double QCDResummation::res_xs(double m, double sqrt_s,
				double muF_m_ratio, double muR_m_ratio,
				log_order LOGo, order match_order, double *asinput) {
    doRapidityDistr = false;
    //
    tau = m*m/sqrt_s/sqrt_s;
    rapY = 1e300;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    //
    LL_order = LOGo;
    matching_order = match_order;
    //
    if(asinput == NULL) as = alphas(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << "  " << muR << "  "<< as << endl;
    //
    bool redo_cheb = !(last_Y==rapY && last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommon(redo_cheb);
    Chebyshev cheb(this);
    Lagrange lagr(this);
    //
    //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    //gsl_set_error_handler_off();
    //
    double result = 0.;

    if(prescr < 0 || expBP) { // then it's a MP
      MP_slope = 1.;
      MP_reN0 = MPbeta + 3.;
      double NL = Nlandau();
      if(MP_reN0 >= NL) {
	MP_reN0 = ( MPbeta + NL )/2.;
	if(MP_reN0 >= NL) {
	  cout << "ERROR: Landau pole N_L = " << NL
	       << " to the left of the rightmost singularity: " << MPbeta << endl;
	  exit(-13);
	}
      }
      double factor = 0.5, freezing = 0.5;
      MPzmin = factor*zminoftau(tau, freezing);
      if(tau>freezing) MPzmin = factor*zminoftau(freezing, freezing);
      if(redo_cheb) {
	if(interpolationMethod==0) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
	if(interpolationMethod==1) lagr.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
      }
      result = MinimalPrescription()/luminosity_in_tau;
    }
    else if(prescr == BP0 || prescr == BP00) {
      if(redo_cheb) cheb.ApproxLumPlus_log1z(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    else {
      if(redo_cheb) cheb.ApproxLumPlus(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    //gsl_set_error_handler (old_handler);
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_Y      = rapY;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = Hmtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }


  double QCDResummation::res_xs_rap(double m, double Y, double sqrt_s,
				    double muF_m_ratio, double muR_m_ratio,
				    log_order LOGo, order match_order, double *asinput) {
    doRapidityDistr = true;
    //
    rapY = Y;
    tau = m*m/sqrt_s/sqrt_s;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    //
    LL_order = LOGo;
    matching_order = match_order;
    //
    if(asinput == NULL) as = alphas(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << as << endl;
    //
    bool redo_cheb = !(last_Y==rapY && last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommon(redo_cheb);
    Chebyshev cheb(this);
    Lagrange lagr(this);
    //
    //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    //gsl_set_error_handler_off();
    //
    double result = 0.;

    if(prescr < 0 || expBP) { // then it's a MP
      MP_slope = 1.;
      MP_reN0 = MPbeta + 3.;
      //if(tau > 0.01) MP_reN0 += 2.;
      double NL = Nlandau();
      if(MP_reN0 >= NL) {
	MP_reN0 = ( MPbeta + NL )/2.;
	if(MP_reN0 >= NL) {
	  cout << "ERROR: Landau pole N_L = " << NL
	       << " to the left of the rightmost singularity: " << MPbeta << endl;
	  exit(-13);
	}
      }
      double factor = 0.1, freezing = 0.5;
      MPzmin = factor*zminoftau(tau, freezing);
      if(tau>freezing) MPzmin = factor*zminoftau(freezing, freezing);
      //MPzmin = tau*exp(2*fabs(Y)); // we can use this tighter lower limit
      //if(tau<MPzmin) MPzmin=tau;
      //if(tau>0.01) MPzmin = 0.01;
      //double Ythr = -0.5*log(MPzmin) * 0.7;
      //double Yeff = (fabs(Y) < Ythr ? Y : Y/fabs(Y)*Ythr);
      //MPzmin *= exp(2*fabs(Yeff)); // rescale?
      if(redo_cheb) {
	if(interpolationMethod==0) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
	if(interpolationMethod==1) lagr.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
      }
      result = MinimalPrescription()/luminosity_in_tau;
    }
    else if(prescr == BP0 || prescr == BP00) {
      if(redo_cheb) cheb.ApproxLumPlus_log1z(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    else {
      if(redo_cheb) cheb.ApproxLumPlus(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    //gsl_set_error_handler (old_handler);
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_Y      = rapY;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = Hmtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }





};






